@@ -1,9 +1,9 @@
 import numpy as np
-import pandas as pd
 import numba
 from warnings import warn
 
 try:
+    import pandas as pd
     import datashader as ds
     import datashader.transfer_functions as tf
     import datashader.bundling as bd
@@ -25,11 +25,11 @@ except ImportError:
 
     or via conda using
 
-    conda install datashader bokeh holoviews colorcet
+     conda install pandas matplotlib datashader bokeh holoviews colorcet scikit-image
     """
     )
     raise ImportError(
-        "umap.plot requires matplotlib, datashader, colorcet and holoviews to be "
+        "umap.plot requires pandas matplotlib datashader bokeh holoviews scikit-image and colorcet to be "
         "installed"
     ) from None
 
@@ -39,11 +39,12 @@ import sklearn.neighbors
 
 from matplotlib.patches import Patch
 
-from umap.nndescent import initialise_search, initialized_nnd_search
-from umap.utils import deheap_sort, submatrix
+from umap.utils import submatrix
 
-from bokeh.plotting import output_notebook, output_file
 from bokeh.plotting import show as show_interactive
+from bokeh.plotting import output_file, output_notebook
+from bokeh.layouts import column
+from bokeh.models import CustomJS, TextInput
 from matplotlib.pyplot import show as show_static
 
 from warnings import warn
@@ -148,6 +149,31 @@ _themes = {
 _diagnostic_types = np.array(["pca", "ica", "vq", "local_dim", "neighborhood"])
 
 
+def _get_embedding(umap_object):
+    if hasattr(umap_object, "embedding_"):
+        return umap_object.embedding_
+    elif hasattr(umap_object, "embedding"):
+        return umap_object.embedding
+    else:
+        raise ValueError("Could not find embedding attribute of umap_object")
+
+
+def _get_metric(umap_object):
+    if hasattr(umap_object, "metric"):
+        return umap_object.metric
+    else:
+        # Assume euclidean if no attribute per cuML.UMAP
+        return "euclidean"
+
+
+def _get_metric_kwds(umap_object):
+    if hasattr(umap_object, "_metric_kwds"):
+        return umap_object._metric_kwds
+    else:
+        # Assume no keywords exist
+        return {}
+
+
 def _embed_datashader_in_an_axis(datashader_image, ax):
     img_rev = datashader_image.data[::-1]
     mpl_img = np.dstack([_blue(img_rev), _green(img_rev), _red(img_rev)])
@@ -156,7 +182,7 @@ def _embed_datashader_in_an_axis(datashader_image, ax):
 
 
 def _nhood_search(umap_object, nhood_size):
-    if umap_object._small_data:
+    if hasattr(umap_object, "_small_data") and umap_object._small_data:
         dmat = sklearn.metrics.pairwise_distances(umap_object._raw_data)
         indices = np.argpartition(dmat, nhood_size)[:, :nhood_size]
         dmat_shortened = submatrix(dmat, indices, nhood_size)
@@ -166,39 +192,10 @@ def _nhood_search(umap_object, nhood_size):
     else:
         rng_state = np.empty(3, dtype=np.int64)
 
-        if len(umap_object._metric_kwds) >= 1:
-            _dist = umap_object._input_distance_func
-            _args = tuple(umap_object._metric_kwds.values())
-
-            @numba.njit()
-            def _metric(x, y):
-                _dist(x, y, *_args)
-
-        else:
-            _metric = umap_object._input_distance_func
-
-        init = initialise_search(
-            umap_object._rp_forest,
-            umap_object._raw_data,
-            umap_object._raw_data,
-            int(nhood_size * umap_object.transform_queue_size),
-            rng_state,
-            _metric,
+        indices, dists = umap_object._knn_search_index.query(
+            umap_object._raw_data, k=nhood_size,
         )
 
-        result = initialized_nnd_search(
-            umap_object._raw_data,
-            umap_object._search_graph.indptr,
-            umap_object._search_graph.indices,
-            init,
-            umap_object._raw_data,
-            _metric,
-        )
-
-        indices, dists = deheap_sort(result)
-        indices = indices[:, :nhood_size]
-        dists = dists[:, :nhood_size]
-
     return indices, dists
 
 
@@ -408,7 +405,10 @@ def _matplotlib_points(
                     "Color key must have enough colors for the number of labels"
                 )
 
-            new_color_key = {k: color_key[i] for i, k in enumerate(unique_labels)}
+            new_color_key = {
+                k: matplotlib.colors.to_hex(color_key[i])
+                for i, k in enumerate(unique_labels)
+            }
             legend_elements = [
                 Patch(facecolor=color_key[i], label=k)
                 for i, k in enumerate(unique_labels)
@@ -475,6 +475,7 @@ def points(
     height=800,
     show_legend=True,
     subset_points=None,
+    ax=None,
 ):
     """Plot an embedding as points. Currently this only works
     for 2D embeddings. While there are many optional parameters
@@ -574,17 +575,21 @@ def points(
         A way to select a subset of points based on an array of boolean
         values.
 
+    ax: matplotlib axis (optional, default None)
+        The matplotlib axis to draw the plot to, or if None, which is
+        the default, a new axis will be created and returned.
+
     Returns
     -------
     result: matplotlib axis
         The result is a matplotlib axis with the relevant plot displayed.
-        If you are using a notbooks and have ``%matplotlib inline`` set
+        If you are using a notebooks and have ``%matplotlib inline`` set
         then this will simply display inline.
     """
-    if not hasattr(umap_object, "embedding_"):
-        raise ValueError(
-            "UMAP object must perform fit on data before it can be visualized"
-        )
+    # if not hasattr(umap_object, "embedding_"):
+    #     raise ValueError(
+    #         "UMAP object must perform fit on data before it can be visualized"
+    #     )
 
     if theme is not None:
         cmap = _themes[theme]["cmap"]
@@ -596,7 +601,7 @@ def points(
             "Conflicting options; only one of labels or values should be set"
         )
 
-    points = umap_object.embedding_
+    points = _get_embedding(umap_object)
 
     if subset_points is not None:
         if len(subset_points) != points.shape[0]:
@@ -617,9 +622,10 @@ def points(
 
     font_color = _select_font_color(background)
 
-    dpi = plt.rcParams["figure.dpi"]
-    fig = plt.figure(figsize=(width / dpi, height / dpi))
-    ax = fig.add_subplot(111)
+    if ax is None:
+        dpi = plt.rcParams["figure.dpi"]
+        fig = plt.figure(figsize=(width / dpi, height / dpi))
+        ax = fig.add_subplot(111)
 
     if points.shape[0] <= width * height // 10:
         ax = _matplotlib_points(
@@ -651,12 +657,12 @@ def points(
         )
 
     ax.set(xticks=[], yticks=[])
-    if umap_object.metric != "euclidean":
+    if _get_metric(umap_object) != "euclidean":
         ax.text(
             0.99,
             0.01,
             "UMAP: metric={}, n_neighbors={}, min_dist={}".format(
-                umap_object.metric, umap_object.n_neighbors, umap_object.min_dist
+                _get_metric(umap_object), umap_object.n_neighbors, umap_object.min_dist
             ),
             transform=ax.transAxes,
             horizontalalignment="right",
@@ -812,7 +818,7 @@ def connectivity(
         edge_cmap = _themes[theme]["edge_cmap"]
         background = _themes[theme]["background"]
 
-    points = umap_object.embedding_
+    points = _get_embedding(umap_object)
     point_df = pd.DataFrame(points, columns=("x", "y"))
 
     point_size = 100.0 / np.sqrt(points.shape[0])
@@ -986,7 +992,7 @@ def diagnostic(
         then this will simply display inline.
     """
 
-    points = umap_object.embedding_
+    points = _get_embedding(umap_object)
 
     if points.shape[1] != 2:
         raise ValueError("Plotting is currently only implemented for 2D embeddings")
@@ -1184,6 +1190,9 @@ def interactive(
     height=800,
     point_size=None,
     subset_points=None,
+    interactive_text_search=False,
+    interactive_text_search_columns=None,
+    interactive_text_search_alpha_contrast=0.95,
 ):
     """Create an interactive bokeh plot of a UMAP embedding.
     While static plots are useful, sometimes a plot that
@@ -1290,6 +1299,16 @@ def interactive(
         A way to select a subset of points based on an array of boolean
         values.
 
+    interactive_text_search: bool (optional, default False)
+        Whether to include a text search widget above the interactive plot
+
+    interactive_text_search_columns: list (optional, default None)
+        Columns of data source to search. Searches labels and hover_data by default.
+
+    interactive_text_search_alpha_contrast: float (optional, default 0.95)
+        Alpha value for points matching text search. Alpha value for points
+        not matching text search will be 1 - interactive_text_search_alpha_contrast
+
     Returns
     -------
 
@@ -1304,7 +1323,7 @@ def interactive(
             "Conflicting options; only one of labels or values should be set"
         )
 
-    points = umap_object.embedding_
+    points = _get_embedding(umap_object)
     if subset_points is not None:
         if len(subset_points) != points.shape[0]:
             raise ValueError(
@@ -1320,7 +1339,7 @@ def interactive(
     if point_size is None:
         point_size = 100.0 / np.sqrt(points.shape[0])
 
-    data = pd.DataFrame(umap_object.embedding_, columns=("x", "y"))
+    data = pd.DataFrame(_get_embedding(umap_object), columns=("x", "y"))
 
     if labels is not None:
         data["label"] = labels
@@ -1367,11 +1386,13 @@ def interactive(
             tooltip_dict = {}
             for col_name in hover_data:
                 data[col_name] = hover_data[col_name]
-                tooltip_dict[col_name] = "@" + col_name
+                tooltip_dict[col_name] = "@{" + col_name + "}"
             tooltips = list(tooltip_dict.items())
         else:
             tooltips = None
 
+        data["alpha"] = 1
+
         # bpl.output_notebook(hide_banner=True) # this doesn't work for non-notebook use
         data_source = bpl.ColumnDataSource(data)
 
@@ -1381,11 +1402,75 @@ def interactive(
             tooltips=tooltips,
             background_fill_color=background,
         )
-        plot.circle(x="x", y="y", source=data_source, color=colors, size=point_size)
+        plot.circle(
+            x="x",
+            y="y",
+            source=data_source,
+            color=colors,
+            size=point_size,
+            alpha="alpha",
+        )
 
         plot.grid.visible = False
         plot.axis.visible = False
 
+        if interactive_text_search:
+            text_input = TextInput(value="", title="Search:")
+
+            if interactive_text_search_columns is None:
+                interactive_text_search_columns = []
+                if hover_data is not None:
+                    interactive_text_search_columns.extend(hover_data.columns)
+                if labels is not None:
+                    interactive_text_search_columns.append("label")
+
+            if len(interactive_text_search_columns) == 0:
+                warn(
+                    "interactive_text_search_columns set to True, but no hover_data or labels provided."
+                    "Please provide hover_data or labels to use interactive text search."
+                )
+
+            else:
+                callback = CustomJS(
+                    args=dict(
+                        source=data_source,
+                        matching_alpha=interactive_text_search_alpha_contrast,
+                        non_matching_alpha=1 - interactive_text_search_alpha_contrast,
+                        search_columns=interactive_text_search_columns,
+                    ),
+                    code="""
+                    var data = source.data;
+                    var text_search = cb_obj.value;
+                    
+                    var search_columns_dict = {}
+                    for (var col in search_columns){
+                        search_columns_dict[col] = search_columns[col]
+                    }
+                    
+                    // Loop over columns and values
+                    // If there is no match for any column for a given row, change the alpha value
+                    var string_match = false;
+                    for (var i = 0; i < data.x.length; i++) {
+                        string_match = false
+                        for (var j in search_columns_dict) {
+                            if (String(data[search_columns_dict[j]][i]).includes(text_search) ) {
+                                string_match = true
+                            }
+                        }
+                        if (string_match){
+                            data['alpha'][i] = matching_alpha
+                        }else{
+                            data['alpha'][i] = non_matching_alpha
+                        }
+                    }
+                    source.change.emit();
+                """,
+                )
+
+                text_input.js_on_change("value", callback)
+
+                plot = column(text_input, plot)
+
         # bpl.show(plot)
     else:
         if hover_data is not None:
@@ -1393,6 +1478,10 @@ def interactive(
                 "Too many points for hover data -- tooltips will not"
                 "be displayed. Sorry; try subssampling your data."
             )
+        if interactive_text_search:
+            warn(
+                "Too many points for text search." "Sorry; try subssampling your data."
+            )
         hv.extension("bokeh")
         hv.output(size=300)
         hv.opts('RGB [bgcolor="{}", xaxis=None, yaxis=None]'.format(background))
